{{
Copyright (c) 2009 Michael Green.  See end of file for terms of use.
}}

'' Once a valid filename is given, the text in the file is copied to SRAM.
'' The format is doubly linked with the first link the reverse pointer and
'' the second link the forward pointer with a zero pointer as the null value.
'' The first pair of links in SRAM are to the last and first lines of text.
'' The next link in SRAM points to the beginning of the free list (of lines).
'' To avoid the need for garbage collection, all lines are fixed length with
'' an effective length byte followed by room for dsp#cols characters.

'' Once the SRAM is loaded, the first dsp#rows lines of the file are displayed
'' and the user is allowed to edit the display buffer.

OBJ
   dsp  : "vga_Driver"                 ' VGA Text Driver
   key  : "comboKeyboard"              ' Keyboard Driver
   sto  : "massStorage"                ' EEPROM, SRAM, and Winbond Flash

CON
   version   = 1                       ' Major version
   release   = 0                       ' Minor release
   testLevel = 2                       ' Test change level
   
   _clkmode  = xtal1 + pll16x
   _xinfreq  = 5_000_000

   sram0CS   =  0                      ' SRAM 0 CS pin number
   sram1CS   =  1                      ' SRAM 1 CS pin number
   winPinDO  =  4                      ' Winbond & SRAM DO pin
   winPinClk =  5                      ' Winbond & SRAM Clk pin
   winPinDI  =  6                      ' Winbond & SRAM DI pin
   winPinCS  =  7                      ' Winbond CS pin number
   tvPins    = 12                      ' TV pin numbers
   vgaPins   = 16                      ' VGA pin numbers
   keyPins   = 26                      ' Normal PS/2 keyboard

   bspKey    = $C8                     ' PS/2 keyboard backspace key
   breakKey  = $CC                     ' PS/2 keyboard application key
      
   lineLen   = dsp#cols                ' Maximum input line length
   endFile   = $FF                     ' Erased byte - end of file mark
   quote     = 34                      ' Double quote
   tabCols   =  8                      ' Number of columns per tab stop
   filesLine = dsp#cols / 14           ' Number of file names on a line
   numSize   =  3                      ' Size of line number field
   msgArea   = dsp#cols - 30           ' Start of error message area
   cmdLine   = dsp#rows - 1            ' Command line (and error message)
   cmdArea   =  2                      ' Start of command entry area
   barSize   = 25                      ' Length of progress bar ("dots")

   startTxt  =  2 * dsp#rows + 8       ' Start of text storage in SRAM
   nullPtr   =  0                      ' Null pointer to text line
   backPtr   =  0                      ' Offset to pointer to previous line
   forwPtr   =  2                      ' Offset to pointer to next line
   textLen   =  4                      ' Offset to text length byte
   textArea  =  5                      ' Offset to text storage (dsp#cols)

   txtFile   =  "t" | "x"<<8 | "t"<<16 ' Extension for text files
   binFile   =  "b" | "i"<<8 | "n"<<16 ' Extension for binary files
   extMask   =  $FF | $FF<<8 | $FF<<16 ' Empty extension

VAR
   long ioControl[2], fcb[4]
   word freeLines, firstNum, firstLine, lastLine, dispRows[dsp#rows]
   byte scanPtr, scanLine[lineLen+1]

PUB main | err, s, state
'' Clear the program space and variables, then read a line and interpret it.
   dsp.start(vgaPins)                  ' Start display
   sto.winSetVideo(dsp.videoCog(0),dsp.videoCog(1))
   key.start(keyPins)                  ' Start keyboard driver
   sto.start(@ioControl)               ' Start I2C/SPI driver
   waitcnt(clkfreq + cnt)
   dsp.str(string("Ouroboros Editor"))
   if version > 0 or release > 0 or testLevel > 0
     dsp.str(string(" Version "))
     dsp.dec(version)
     dsp.out(".")
     if release < 100
       dsp.out("0")
     if release < 10
       dsp.out("0")
     dsp.dec(release)
     if testLevel > 0
       dsp.out("a"+testLevel-1)
   dsp.str(string(dsp#Cr,dsp#Lf))
   dsp.str(string("Free space: "))
   dsp.dec((dsp#sync - @s - 4)>>10)
   dsp.str(string("K",dsp#Cr,dsp#Lf))
   dsp.str(string("Mounting flash ... "))
   dsp.dec(sto.winStart(winPinCS,winPinClk,winPinDI,winPinDO,sram0CS,sram1CS)*4)
   dsp.str(string("K free out of "))
   dsp.dec(sto.flashSize>>20)
   dsp.str(string("MB",dsp#Cr,dsp#Lf,dsp#Esc,"[3z"))
   initializeSRAM
   waitcnt(clkfreq + cnt)
   if not key.present
      dsp.str(string("No keyboard present",dsp#Cr))
      abort
   key.clearkeys
   listFiles
   repeat
      dsp.setLine(msgArea,cmdLine,string("Enter a file name"))
      askYesNo                         ' Get entered command
      ifnot scanLine[0]
         listFiles                     ' No file name?  List directory
         next
      if sto.initFile(@fcb,@scanLine)
         if fcb[2] & extMask == binFile' Extension given was ".bin"
            ifnot sto.openFile(@fcb)   ' Try to open the file name
               dsp.setLine(msgArea,cmdLine,string("Can't open executable"))
               next
            sto.bootFile(@fcb)         ' Execute the file with Spin
         if fcb[2] & extMask == extMask' No extension was provided
            if sto.openFile(@fcb)      ' Try to open the file name
               sto.bootFile(@fcb)      ' Assume it's an executable
            fcb[2] := binFile          ' Can't open, try XXX.bin
            if sto.openFile(@fcb)
               sto.bootFile(@fcb)      ' Execute the file if present
            fcb[2] := txtFile          ' Try XXX.txt also
         if sto.openFile(@fcb)
            quit
         else
            dsp.setLine(msgArea,cmdLine,string("New file.  Create? (y/N)"))
            if askYesNo
               quit
      else
         dsp.setLine(msgArea,cmdLine,string("Invalid filename"))
   if (s := copyToSRAM(@fcb)) <> 0     ' Error copying file
      dsp.setLine(msgArea,cmdLine,s)
   else
      copyToDisplay(0,dsp#rows-1)
      fileEditor
      copyFromDisplay(0,dsp#rows-1)
      if (s := copyFromSRAM(@fcb)) <> 0' Error copying file
         dsp.setLine(msgArea,cmdLine,s)
      else
         dsp.setLine(msgArea,cmdLine,string("Editing complete"))
   sto.writeSRAM(0,@freeLines,startTxt)' Write debugging info
   waitcnt(clkfreq + cnt)              ' Pause for a second
   if sto.initFile(@fcb,string("femtobas.bin"))
      if sto.openFile(@fcb)
         sto.bootFile(@fcb)
   reboot

PRI askYesNo | state                   ' Allow user to enter text
   dsp.setCursor(0,cmdLine)            ' Return true if yes/YES/y/Y
   dsp.str(string(">>"))               ' Return false otherwise
   state~
   repeat
      state := windowEditor(cmdArea,cmdLine,msgArea-cmdArea-1,1,state)
   until state & $FF == $0D
   dsp.clearRegion(msgArea,cmdLine,30,1) ' Clear message area & get text
   dsp.getLine(@scanLine,cmdArea,cmdLine,msgArea-cmdArea-1)
   if strcomp(@scanLine,string("yes")) or strcomp(@scanLine,string("y"))
      return true
   if strcomp(@scanLine,string("YES")) or strcomp(@scanLine,string("Y"))
      return true

PRI listFiles | a, width, c            ' Display flash directory
   dsp.clearRegion(0,0,dsp#cols,cmdLine) ' Clear editing area
   dsp.setCursor(0,0)
   dsp.str(string("Current Flash Files",dsp#Cr,dsp#Lf,dsp#Lf))
   sto.firstFile(@fcb)
   c := filesLine                      ' 5 file names per line = 70 cols
   repeat while sto.nextFile(@fcb)
      width := 2                       ' 2 spaces between file names
      repeat a from 0 to 7
         if fcb.byte[a] == $FF         ' Trailing pad for $FF
            width++
         else
            dsp.out(fcb.byte[a])
      if fcb[2] & $00FFFFFF == $00FFFFFF
         width++                       ' Trailing pad for no "."
      else
         dsp.out(".")
      repeat a from 8 to 10
         if fcb.byte[a] == $FF         ' Trailing pad for $FF
            width++
         else
            dsp.out(fcb.byte[a])
      if --c                           ' Don't pad after last file name
         repeat width
            dsp.out(" ")
      else
         dsp.str(string(dsp#Cr,dsp#Lf))
         c := filesLine                ' Start a new display line
   if c <> filesLine
      dsp.str(string(dsp#Cr,dsp#Lf))   ' Finish any partial line

PRI initializeSRAM | n, c, s, d        ' Initialize SRAM
   d := 0
   dsp.str(string("Initializing SRAM ...      "))
   repeat c from 0 to dsp#rows-1       ' Display buffer
      dispRows[c] := nullPtr
   s := dsp#cols + 5                   ' Used for each line
   firstLine := lastLine := nullPtr    ' Empty text list
   firstNum := 0
   freeLines := c := startTxt          ' Free list
   n := c + s
   repeat until c == nullPtr
      dsp.setCol(22)
      dsp.dec(++d)
      sto.writeSRAM(c+forwPtr,n,-2)    ' Link to next line
      c := n
      n += s
      if n > $8000 - s and n < $8000   ' SRAM data can't span SRAMs
         n := $8000
      if n > $10000 - s                ' No data beyond end of SRAM
         n := nullPtr
   dsp.str(string(" lines",dsp#Cr,dsp#Lf))

PRI copyToSRAM(p) | i, n, c, r, t, e, s' Copy opened file to SRAM
   c := n := r := 0                    '  p points to FCB
   firstNum := 1
   e := long[p][2] >> 24 < $FF         ' Valid file segment
   repeat while e or r < dsp#rows      ' Display progress bar
      if n // barSize == 0
         dsp.clearRegion(0,dsp#rows-1,barSize,1)
      dsp.setLine(n++//barSize,dsp#rows-1,string("."))
      scanPtr := 0
      if e                             ' If file open, get lines
         repeat
            ifnot sto.readFile(p,@c,1)
               return string("Unexpected end of file")
            case c
               dsp#Lf:  next           ' Skip LF
               dsp#Cr:  quit           ' Marks end of line
               endFile: quit           ' End of file
               other:   if scanPtr < dsp#cols
                           scanLine[scanPtr++] := c
                        else
                           return string("Line too long")
         if c == endFile
            e := false                 ' No more text lines available
      if (i := freeLines) == nullPtr
         return string("Out of memory") ' Get line from free list
      sto.readSRAM(i+forwPtr,@freeLines,2)
      t := lastLine                    ' Get link to last line
      sto.writeSRAM(t+forwPtr,@i,2)    '  and link in new line
      lastLine := i
      if firstLine == nullPtr          ' If list empty, start it
         firstLine := i
      if r < dsp#rows                  ' Save pointers to dsp#rows rows
         dispRows[r++] := i
      sto.writeSRAM(i+backPtr,nullPtr<<16|t,-4) ' Make new line last
      sto.writeSRAM(i+textLen,scanPtr,-1) ' Copy flash line to SRAM
      if scanPtr > 0
         sto.writeSRAM(i+textArea,@scanLine,scanPtr)
   dsp.clearRegion(0,dsp#rows-1,barSize,1)
   return false                        ' File copied to SRAM

PRI copyToDisplay(yy,h) | p, i, num, f ' Copy SRAM lines to display
   repeat i from yy to yy+h-1
      f := dsp.decString(firstNum+i-yy,@num)
      dsp.setLine(0,i,string("    "))  ' Right justify line number in
      dsp.setLine(numSize-f,i,@num)    '  first 3 columns of display
      p := dispRows[i]
      if (scanPtr := sto.readSRAM(p+textLen,0,-1)) > 0
         sto.readSRAM(p+textArea,@scanLine,scanPtr)
      scanLine[scanPtr]~               ' Put string terminator
      dsp.setLine(numSize+1,i,@scanLine)' Copy string to display
      if scanPtr+numSize+1 < dsp#cols  ' Erase rest of line
         dsp.clearRegion(scanPtr+numSize+1,i,dsp#cols-scanPtr-numSize-1,1)

PRI copyFromDisplay(yy,h) | i, p       ' Copy display lines to SRAM
   repeat i from yy to yy+h-1
      dsp.getLine(@scanLine,numSize+1,i,dsp#cols-numSize-1)
      scanPtr := strsize(@scanLine)
      p := dispRows[i]
      sto.writeSRAM(p+textLen,scanPtr,-1) ' Write text to SRAM
      if scanPtr > 0
         sto.writeSRAM(p+textArea,@scanLine,scanPtr)

PRI copyFromSRAM(p) | i, n, e, s       ' Copy a text file from SRAM to
   i := n := 0                         '  the file name in the fcb
   ifnot sto.createFile(p)             ' Try to create the file
      ifnot sto.eraseFile(p)           ' If already exists, erase it
         return string("Internal error (erase)")
      ifnot sto.createFile(p)          '  and try to create it again
         return string("Can't create file")
   if (e := lastLine) <> nullPtr       ' Ignore trailing empty lines
      repeat while sto.readSRAM(e+textLen,0,-1) == 0
         if (e := sto.readSRAM(e+backPtr,0,-2)) == nullPtr
            quit
   i := firstLine                      ' Start at beginning of SRAM
   repeat until i == nullPtr or e == nullPtr
      if n // barSize == 0
         dsp.clearRegion(0,dsp#rows-1,barSize,1)
      dsp.setLine(n++//barSize,dsp#rows-1,string("."))
      if (scanPtr := sto.readSRAM(i+textLen,0,-1)) > 0
         sto.readSRAM(i+textArea,@scanLine,scanPtr) ' Copy text to file
         sto.writeFile(p,@scanLine,scanPtr)
      sto.writeFile(p,string(dsp#Cr,dsp#Lf),2) ' Add CR/LF
      if i == e                        ' Stop after last non-blank line
         quit
      i := sto.readSRAM(i+forwPtr,0,-2)' Follow forward link
   dsp.clearRegion(0,dsp#rows-1,barSize,1)
   return false

PUB scrollUpFile(yy,h) | i, t, u, p, n ' Scroll upwards one line
   u := yy + ||h - 1                   ' If h < 0, delete the top line
   if h > 0                            ' If h > 0, top line goes to file
      copyFromDisplay(yy,1)            ' Copy top line back to file
      firstNum++                       ' Top line is numbered higher
   else                                ' **** What about last line???
      t := dispRows[yy]                ' ****  (yy == dsp#rows-2)
      p := sto.readSRAM(t+backPtr,0,-2)' Delete the top line
      n := sto.readSRAM(t+forwPtr,0,-2)
      sto.writeSRAM(p+forwPtr,n,-2)    ' Patch pointers
      sto.writeSRAM(n+backPtr,p,-2)
      sto.writeSRAM(t+forwPtr,freeLines,-2)
      freeLines := t                   ' Put deleted line on free list
      if lastLine == t                 ' **** What if at first line???
         lastLine := n
   if ||h > 1
      dsp.rollUpRegion(0,yy,dsp#cols,||h) ' Scroll up display
      repeat i from yy to u-1          ' Shift the pointers upwards
         dispRows[i] := dispRows[i+1]  '  and get next line pointer
   if (t := dispRows[u-1]) == lastLine or lastLine == nullPtr
      if (i := freeLines) == nullPtr   ' At end of file, add new line
         return string("Out of memory")' Get line from free list
      sto.readSRAM(i+forwPtr,@freeLines,2)
      sto.writeSRAM(i+backPtr,lastLine|nullPtr<<16,-4)
      sto.writeSRAM(i+textLen,0,-1)    ' New line is blank
      dispRows[u] := lastLine := i
   else                                ' Not end of file, forwPtr OK
      dispRows[u] := sto.readSRAM(t+forwPtr,0,-2)
   copyToDisplay(u,1)                  ' Copy new line to display
   return false

PUB scrollDownFile(yy,h) | i, t, u, p, n ' Scroll downwards one line
   if h > 0                            ' If height > 0, scroll backwards
      if (t := dispRows[yy]) == firstLine ' Beginning of file
         return string("At beginning of file")
   else                                ' If height < 0, insert blank line
      if (t := freeLines) == nullPtr   ' Insert a new blank line
         return string("Out of memory")' Get line from free list
      sto.readSRAM(t+forwPtr,@freeLines,2)
      n := dispRows[yy]
      p := sto.readSRAM(n+backPtr,0,-2)' Link in new line
      sto.writeSRAM(n+backPtr,t,-2)
      sto.writeSRAM(p+forwPtr,t,-2)
      sto.writeSRAM(t+backPtr,p|n<<16,-4)
      sto.writeSRAM(t+textLen,0,-1)    ' New line is blank
   u := yy + ||h - 1
   copyFromDisplay(u,1)                ' Save bottom of display
   if ||h > 1
      dsp.rollDnRegion(0,yy,dsp#cols,h)' Scroll down display
      repeat i from u-1 to yy          ' Shift the pointers downwards
         dispRows[i+1] := dispRows[i]  '  and get next line pointer
   if h > 0
      dispRows[yy] := sto.readSRAM(t+backPtr,0,-2)
      firstNum--
   else
      dispRows[yy] := t
   copyToDisplay(yy,1)                 ' Copy new line to display
   return false

PUB windowEditor(winX, winY, winC, winR, state) | c, t, relCol, relRow

' This method provides an editable window on the display text screen.
' It uses the standard PS/2 keyboard control keys including backspace,
' cursor positioning, delete and insert (for an insert mode).
' The edited text must be retrieved from the window by the caller.
' Unknown or invalid control keys terminate editing and return the keycode.
' The state of the window and the screen is such that editing can be
' resumed after the terminating keycode is processed by the caller.
' The return value contains: insert << 31 | y << 16 | x << 8 | keycode
' and is normally used on a subsequent call to maintain the state.
' A value of zero is used to initialize the editor to the upper left
' corner of the window.  No checking is done for valid window bounds.

   if state == 0
      dsp.clearRegion(winX,winY,winC,winR)
   dsp.setCursor(((state>>8)&$FF)+winX,((state>>16)&$FF)+winY)
   if state ~>= 31
      dsp.cursorMode(dsp#fastCurs,true)
   else
      dsp.cursorMode(dsp#slowCurs,true)
   repeat
      relCol := dsp.getCol - winX
      relRow := dsp.getRow - winY
      case c := key.getkey
         $09:                          ' Tab
            repeat
               if state                ' Insert spaces to next tab stop
                  dsp.insertSpaces(dsp.getCol,dsp.getRow,1)
               dsp.out(" ")
            while (dsp.getCol - winX) // tabCols
         $0D:                          ' Return
            if state                   ' Insert mode
               return state<<31 | relRow<<16 | relCol<<8 | c
            else                       ' Normal mode
               dsp.clearRegion(dsp.getCol,dsp.getRow,winC-relCol,1)
               dsp.setCol(winX)        ' Go to beginning of next line
               relCol := dsp.getCol - winX
               relRow := dsp.getRow - winY
               if relRow == winR-1     ' Exit if at bottom of window
                  return state<<31 | relRow<<16 | relCol<<8 | c
               dsp.out(dsp#Lf)
         $20..$7E:                     ' Printable characters
            if relCol < winC-1
               if state
                  dsp.insertSpaces(winX+relCol,winY+relRow,1)
               dsp.out(c)
         $C0:                          ' Left arrow
            if relCol > 0
               dsp.str(string(dsp#Esc,"[D"))
         $C1:                          ' Right arrow
            if relCol < winC-1
               dsp.str(string(dsp#Esc,"[C"))
         $C2:                          ' Up arrow
            if relRow == 0             ' Exit if already at top of window
               return state<<31 | relRow<<16 | relCol<<8 | c
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.str(string(dsp#Esc,"[A"))
         $C3:                          ' Down arrow
            if relRow == winR-1        ' Exit if already at bottom of window
               return state<<31 | relRow<<16 | relCol<<8 | c
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.str(string(dsp#Esc,"[B"))
         $C4:                          ' Home
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.setCursor(0,0)
         $C5:                          ' End
            if winR == 1               ' Special case for a single line window
               return state<<31 | relRow<<16 | relCol<<8 | c
            dsp.setCursor(0,winY+winR-1)
         $C8:                          ' Backspace
            if relCol > 0              ' Not at left end of line
               dsp.deleteText(winX+relCol-1,winY+relRow,1)
               dsp.setCol(winX+relCol-1)
            else                       ' Already at left end of line
               return state<<31 | relRow<<16 | relCol<<8 | c
         $C9, $7E:                     ' Delete
            if dsp.getLine(0,winX+relCol,winY+relRow,dsp#cols-winX-relCol) == 0
               return state<<31 | relRow<<16 | relCol<< 8 | c
            else                       ' Non-blank characters remain to delete
               dsp.deleteText(winX+relCol,winY+relRow,1)
         $CA:                          ' Toggle Insert mode
            if state := not state
               dsp.cursorMode(dsp#fastCurs,true)
            else
               dsp.cursorMode(dsp#slowCurs,true)
         other:                        ' All others, exit
            return state<<31 | relRow<<16 | relCol<<8 | c

PRI fileEditor | c, xx, yy, state      ' Edit the file opened as fcbIn
   c := $CC                            ' Initial editor state
   repeat
      c := windowEditor(numSize+1,0,dsp#cols-numSize-1,dsp#rows-1,c)
      xx := (c >> 8 & $FF) + numSize + 1
      yy := c >> 16 & $FF
      state := c ~> 31
      case c & $FF
         $0D:                          ' Carriage return
            if state                   ' Insert mode at any point
               scanPtr := dsp.getLine(@scanLine,xx,yy,dsp#cols-xx-numSize-1)
               if scanPtr > 0
                  dsp.clearRegion(xx,yy,dsp#cols-xx-numSize-1,1)
               if yy < dsp#rows-2      ' Not at end of display area
                  c := state<<31 | yy++<<16 | (xx-numSize+1)<< 8 | $0D
                  if (result := scrollDownFile(yy,-dsp#rows+yy+1)) <> false
                     dsp.setLine(msgArea,cmdLine,result)
                     quit
                  dsp.setLine(0,yy,@scanLine)
               else
               ' **** Use scrollUpFile if at last line, but need
               ' **** special mode to provide a blank line instead
               ' **** of reading a new line from the file.
            else                       ' Normal mode at bottom of screen
               if (result := scrollUpFile(0,dsp#rows-1)) <> false
                  dsp.setLine(msgArea,cmdLine,result)
                  quit                 ' Out of memory
         $C2:                          ' Up arrow at top of screen
            if (result := scrollDownFile(0,dsp#rows-1)) <> false
               dsp.setLine(msgArea,cmdLine,result)
               quit                    ' Already at beginning of file
         $C3:                          ' Down arrow at bottom of screen
            if (result := scrollUpFile(0,dsp#rows-1)) <> false
               dsp.setLine(msgArea,cmdLine,result)
               quit                    ' Out of memory
         $C6:                          ' PgUp key
            repeat dsp#rows-1
               if (result := scrollDownFile(0,dsp#rows-1)) <> false
                  dsp.setLine(msgArea,cmdLine,result)
                  quit                 ' Already at beginning of file
            if result <> false
               quit
         $C7:                          ' PgDn key
            repeat dsp#rows-1
               if (result := scrollUpFile(0,dsp#rows-1)) <> false
                  dsp.setLine(msgArea,cmdLine,result)
                  quit                 ' Out of memory
         $C8:                          ' Backspace key
            next
         $C9, $7E:                     ' Delete key
            if (result := scrollUpFile(yy,-dsp#rows+yy+1)) <> false
               dsp.setLine(msgArea,cmdLine,result)
               quit
         $CC:                          ' Application key
            dsp.cursorMode(dsp#slowCurs,true)
            return false
         other:                        ' Any unrecognized non-printable
            next

{{
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}

